@@ -2,47 +2,60 @@ import traceback
 import uuid
 from multiprocessing import TimeoutError
 
-from cloudshell.cp.aws.domain.common.exceptions import CancellationException
+from cloudshell.cp.aws.domain.common.cancellation_service import CommandCancellationService
+from cloudshell.cp.aws.domain.common.list_helper import first_or_default
+from cloudshell.cp.aws.domain.common.vm_details_provider import VmDetailsProvider
 from cloudshell.cp.aws.domain.services.ec2.security_group import SecurityGroupService
-from cloudshell.cp.aws.domain.services.ec2.tags import IsolationTagValues
+from cloudshell.cp.aws.domain.services.ec2.tags import IsolationTagValues, TypeTagValues
+from cloudshell.cp.aws.domain.services.ec2.elastic_ip import ElasticIpService
 from cloudshell.cp.aws.domain.services.parsers.port_group_attribute_parser import PortGroupAttributeParser
+from cloudshell.cp.aws.domain.services.strategy.device_index import *
 from cloudshell.cp.aws.models.ami_deployment_model import AMIDeploymentModel
 from cloudshell.cp.aws.models.deploy_result_model import DeployResult
+from cloudshell.cp.aws.models.deploy_aws_ec2_ami_instance_resource_model import DeployAWSEc2AMIInstanceResourceModel
 from cloudshell.shell.core.driver_context import CancellationContext
+from cloudshell.cp.aws.domain.services.ec2.instance import InstanceService
+from cloudshell.cp.aws.domain.services.ec2.instance_credentials import InstanceCredentialsService
+from cloudshell.cp.aws.domain.services.ec2.tags import TagService
+from cloudshell.cp.aws.domain.services.ec2.keypair import KeyPairService
+from cloudshell.cp.aws.domain.services.ec2.vpc import VPCService
+from cloudshell.cp.aws.domain.services.ec2.subnet import SubnetService
+from cloudshell.cp.aws.domain.services.ec2.network_interface import NetworkInterfaceService
+from cloudshell.cp.aws.models.network_actions_models import *
 
 
 class DeployAMIOperation(object):
     MAX_IO1_IOPS = 20000
 
     def __init__(self, instance_service, ami_credential_service, security_group_service, tag_service,
-                 vpc_service, key_pair_service, subnet_service, cancellation_service):
+                 vpc_service, key_pair_service, subnet_service, elastic_ip_service, network_interface_service,
+                 cancellation_service, device_index_strategy, vm_details_provider):
         """
-        :param instance_service: Instance Service
-        :type instance_service: cloudshell.cp.aws.domain.services.ec2.instance.InstanceService
-        :param ami_credential_service: AMI Credential Service
-        :type ami_credential_service: cloudshell.cp.aws.domain.services.ec2.instance_credentials.InstanceCredentialsService
-        :param security_group_service: Security Group Service
-        :type security_group_service: cloudshell.cp.aws.domain.services.ec2.security_group.SecurityGroupService
-        :param tag_service: Tag service
-        :type tag_service: cloudshell.cp.aws.domain.services.ec2.tags.TagService
-        :param vpc_service: VPC service
-        :type vpc_service: cloudshell.cp.aws.domain.services.ec2.vpc.VPCService
-        :param key_pair_service: Key Pair Service
-        :type key_pair_service: cloudshell.cp.aws.domain.services.ec2.keypair.KeyPairService
-        :param subnet_service: Subnet Service
-        :type subnet_service: cloudshell.cp.aws.domain.services.ec2.subnet.SubnetService
-        :param cancellation_service:
-        :type cancellation_service: cloudshell.cp.aws.domain.common.cancellation_service.CommandCancellationService
+        :param InstanceService instance_service: Instance Service
+        :param InstanceCredentialsService ami_credential_service: AMI Credential Service
+        :param SecurityGroupService security_group_service: Security Group Service
+        :param TagService tag_service: Tag service
+        :param VPCService vpc_service: VPC service
+        :param KeyPairService key_pair_service: Key Pair Service
+        :param SubnetService subnet_service: Subnet Service
+        :param ElasticIpService elastic_ip_service: Elastic Ips Service
+        :param NetworkInterfaceService network_interface_service:
+        :param CommandCancellationService cancellation_service:
+        :param AbstractDeviceIndexStrategy device_index_strategy:
+        :param VmDetailsProvider vm_details_provider:
         """
-
         self.tag_service = tag_service
         self.instance_service = instance_service
         self.security_group_service = security_group_service
         self.credentials_service = ami_credential_service
         self.vpc_service = vpc_service
         self.key_pair_service = key_pair_service
-        self.subnet_serivce = subnet_service
+        self.subnet_service = subnet_service
         self.cancellation_service = cancellation_service
+        self.elastic_ip_service = elastic_ip_service
+        self.network_interface_service = network_interface_service
+        self.device_index_strategy = device_index_strategy
+        self.vm_details_provider = vm_details_provider
 
     def deploy(self, ec2_session, s3_session, name, reservation, aws_ec2_cp_resource_model,
                ami_deployment_model, ec2_client, cancellation_context, logger):
@@ -76,12 +89,13 @@ class DeployAMIOperation(object):
 
         instance = None
         security_group = None
-        allocated_elastic_ip = None
+        network_config_results = self._prepare_network_result_models(ami_deployment_model)
         try:
             security_group = self._create_security_group_for_instance(ami_deployment_model=ami_deployment_model,
                                                                       ec2_session=ec2_session,
                                                                       reservation=reservation,
-                                                                      vpc=vpc)
+                                                                      vpc=vpc,
+                                                                      logger=logger)
 
             self.cancellation_service.check_if_cancelled(cancellation_context)
 
@@ -91,7 +105,10 @@ class DeployAMIOperation(object):
                                                                      vpc=vpc,
                                                                      security_group=security_group,
                                                                      key_pair=key_name,
-                                                                     reservation=reservation)
+                                                                     reservation=reservation,
+                                                                     network_config_results=network_config_results,
+                                                                     logger=logger)
+
             instance = self.instance_service.create_instance(ec2_session=ec2_session,
                                                              name=name,
                                                              reservation=reservation,
@@ -101,11 +118,18 @@ class DeployAMIOperation(object):
                                                              cancellation_context=cancellation_context,
                                                              logger=logger)
 
+            logger.info("Instance created, populating results with interface data")
+            self._populate_network_config_results_with_interface_data(instance=instance,
+                                                                      network_config_results=network_config_results)
+
             self.cancellation_service.check_if_cancelled(cancellation_context)
 
-            if ami_deployment_model.allocate_elastic_ip:
-                allocated_elastic_ip = self._set_elastic_ip(ec2_session=ec2_session, ec2_client=ec2_client,
-                                                            instance=instance)
+            self.elastic_ip_service.set_elastic_ips(ec2_session=ec2_session,
+                                                    ec2_client=ec2_client,
+                                                    instance=instance,
+                                                    ami_deployment_model=ami_deployment_model,
+                                                    network_config_results=network_config_results,
+                                                    logger=logger)
 
             self.cancellation_service.check_if_cancelled(cancellation_context)
 
@@ -113,10 +137,11 @@ class DeployAMIOperation(object):
             self._rollback_deploy(ec2_session=ec2_session,
                                   instance_id=self._extract_instance_id_on_cancellation(e, instance),
                                   custom_security_group=security_group,
-                                  elastic_ip=allocated_elastic_ip,
+                                  network_config_results=network_config_results,
                                   logger=logger)
             raise  # re-raise original exception after rollback
 
+        logger.info("Instance {} created, getting ami credentials".format(instance.id))
         ami_credentials = self._get_ami_credentials(key_pair_location=aws_ec2_cp_resource_model.key_pairs_location,
                                                     wait_for_credentials=ami_deployment_model.wait_for_credentials,
                                                     instance=instance,
@@ -126,8 +151,17 @@ class DeployAMIOperation(object):
                                                     cancellation_context=cancellation_context,
                                                     logger=logger)
 
-        deployed_app_attributes = self._prepare_deployed_app_attributes(instance, ami_credentials, ami_deployment_model,
-                                                                        allocated_elastic_ip)
+        logger.info("Preparing result")
+
+        deployed_app_attributes = self._prepare_deployed_app_attributes(ami_credentials=ami_credentials,
+                                                                        ami_deployment_model=ami_deployment_model,
+                                                                        network_config_results=network_config_results)
+
+        vm_details_data = self.vm_details_provider.create(instance)
+
+        network_actions_results_dtos = \
+            self._prepare_network_config_results_dto(network_config_results=network_config_results,
+                                                     ami_deployment_model=ami_deployment_model)
 
         return DeployResult(vm_name=self._get_name_from_tags(instance),
                             vm_uuid=instance.instance_id,
@@ -137,11 +171,78 @@ class DeployAMIOperation(object):
                             auto_delete=True,
                             autoload=ami_deployment_model.autoload,
                             inbound_ports=ami_deployment_model.inbound_ports,
-                            outbound_ports=ami_deployment_model.outbound_ports,
                             deployed_app_attributes=deployed_app_attributes,
                             deployed_app_address=instance.private_ip_address,
-                            public_ip=instance.public_ip_address if ami_deployment_model.add_public_ip else None,
-                            elastic_ip=allocated_elastic_ip)
+                            public_ip=instance.public_ip_address,
+                            network_configuration_results=network_actions_results_dtos,
+                            vpc_id=vpc.id,
+                            availability_zone=self._get_availability_zone(instance, logger),
+                            vm_details_data=vm_details_data)
+
+    def _get_availability_zone(self, instance, logger):
+        if 'AvailabilityZone' in instance.placement:
+            return instance.placement['AvailabilityZone']
+
+        logger.error("Failed to get AvailabilityZone for instance '{}'".format(instance.instance_id))
+        return None
+
+    def _validate_public_subnet_exist_if_requested_public_or_elastic_ips(self, ami_deployment_model, logger):
+        """
+        :param DeployAWSEc2AMIInstanceResourceModel ami_deployment_model:
+        :param logging.Logger logger:
+        """
+        if ami_deployment_model.add_public_ip or ami_deployment_model.allocate_elastic_ip:
+            connect_subnet_actions = \
+                filter(lambda x: isinstance(x.connection_params, SubnetConnectionParams),
+                       ami_deployment_model.network_configurations)
+
+            if not any(x.connection_params.is_public_subnet() for x in connect_subnet_actions):
+                msg = "Cannot deploy app with elastic or public ip when connected only to private subnets"
+                logger.error(msg)
+                raise ValueError(msg)
+
+    def _prepare_network_config_results_dto(self, network_config_results, ami_deployment_model):
+        """
+        :param list[DeployNetworkingResultModel] network_config_results:
+        :param DeployAWSEc2AMIInstanceResourceModel ami_deployment_model:
+        :return:
+         :rtype" list[DeployNetworkingResultDto]
+        """
+        if not ami_deployment_model.network_configurations:
+            return []  # for the moment if we didnt received a connectivity action we shouldnot return anything
+        return list(map(self._convertDeployNetworkResultModelToDto, network_config_results))
+
+    def _convertDeployNetworkResultModelToDto(self, network_config_result):
+        """
+        :param DeployNetworkingResultModel network_config_result:
+        :rtype: ConnectToSubnetActionResult
+        """
+        import json
+        interface_data_json_str = json.dumps({
+            'interface_id': network_config_result.interface_id,
+            'IP': network_config_result.private_ip,
+            'Public IP': network_config_result.public_ip,
+            'Elastic IP': network_config_result.is_elastic_ip,
+            'MAC Address': network_config_result.mac_address,
+            'Device Index': network_config_result.device_index,
+        })
+        return ConnectToSubnetActionResult(action_id=network_config_result.action_id,
+                                           success=True,
+                                           interface_data=interface_data_json_str)
+
+    def _prepare_network_result_models(self, ami_deployment_model):
+        """
+        :param DeployAWSEc2AMIInstanceResourceModel ami_deployment_model:
+        :rtype: list[DeployNetworkingResultModel]
+        """
+        network_config_results = []
+        if ami_deployment_model.network_configurations is None:
+            network_config_results.append(DeployNetworkingResultModel(''))  # init a result object with empty action id
+        else:
+            for net_config in ami_deployment_model.network_configurations:
+                if isinstance(net_config.connection_params, SubnetConnectionParams):
+                    network_config_results.append(DeployNetworkingResultModel(net_config.id))
+        return network_config_results
 
     def _extract_instance_id_on_cancellation(self, exception, instance):
         """
@@ -151,7 +252,8 @@ class DeployAMIOperation(object):
         """
         instance_id = None
         if exception and hasattr(exception, "data") and exception.data and 'instance_ids' in exception.data:
-            instance_id = exception.data['instance_ids'][0]     # we assume at this point that we are working on a single app
+            instance_id = exception.data['instance_ids'][
+                0]  # we assume at this point that we are working on a single app
         elif instance:
             instance_id = instance.id
         return instance_id
@@ -195,13 +297,16 @@ class DeployAMIOperation(object):
     def _get_name_from_tags(result):
         return [tag['Value'] for tag in result.tags if tag['Key'] == 'Name'][0]
 
-    def _create_security_group_for_instance(self, ami_deployment_model, ec2_session, reservation, vpc):
+    def _create_security_group_for_instance(self, ami_deployment_model, ec2_session, reservation, vpc, logger):
         if not ami_deployment_model.inbound_ports and not ami_deployment_model.outbound_ports:
             return None
 
+        logger.info("Parsing inbound_ports attribute")
+
         inbound_ports = PortGroupAttributeParser.parse_port_group_attribute(ami_deployment_model.inbound_ports)
         outbound_ports = PortGroupAttributeParser.parse_port_group_attribute(ami_deployment_model.outbound_ports)
         if not inbound_ports and not outbound_ports:
+            logger.info("No data found in inbound_ports attribute")
             return None
 
         security_group_name = SecurityGroupService.CLOUDSHELL_CUSTOM_SECURITY_GROUP.format(str(uuid.uuid4()))
@@ -212,7 +317,8 @@ class DeployAMIOperation(object):
 
         tags = self.tag_service.get_security_group_tags(name=security_group_name,
                                                         isolation=IsolationTagValues.Exclusive,
-                                                        reservation=reservation)
+                                                        reservation=reservation,
+                                                        type=TypeTagValues.InboundPorts)
 
         self.tag_service.set_ec2_resource_tags(security_group, tags)
 
@@ -222,10 +328,13 @@ class DeployAMIOperation(object):
         if outbound_ports:
             self.security_group_service.remove_allow_all_outbound_rule(security_group=security_group)
 
+        logger.info("Created security group {0} from inbound_ports attribute: {1}"
+                    .format(security_group.group_id, ami_deployment_model.inbound_ports))
+
         return security_group
 
     def _create_deployment_parameters(self, ec2_session, aws_ec2_resource_model, ami_deployment_model, vpc,
-                                      security_group, key_pair, reservation):
+                                      security_group, key_pair, reservation, network_config_results, logger):
         """
         :param ec2_session:
         :param aws_ec2_resource_model: The resource model of the AMI deployment option
@@ -239,6 +348,9 @@ class DeployAMIOperation(object):
         :type key_pair : str
         :param reservation: reservation model
         :type reservation: cloudshell.cp.aws.models.reservation_model.ReservationModel
+        :param network_config_results: list of network configuration result objects
+        :type network_config_results: list[DeployNetworkingResultModel]
+        :param logging.Logger logger:
         """
         aws_model = AMIDeploymentModel()
         if not ami_deployment_model.aws_ami_id:
@@ -256,26 +368,105 @@ class DeployAMIOperation(object):
                                                                           ami_deployment_model=ami_deployment_model,
                                                                           aws_ec2_resource_model=aws_ec2_resource_model)
         aws_model.aws_key = key_pair
-        aws_model.add_public_ip = ami_deployment_model.add_public_ip
 
-        subnet = self.subnet_serivce.get_subnet_from_vpc(vpc)
-        aws_model.subnet_id = subnet.id
+        security_group_ids = self._get_security_group_param(reservation, security_group, vpc,
+                                                            ami_deployment_model.allow_all_sandbox_traffic)
+        aws_model.security_group_ids = security_group_ids
 
-        self._set_security_group_param(aws_model, reservation, security_group, vpc)
+        aws_model.network_interfaces = \
+            self._prepare_network_interfaces(vpc=vpc,
+                                             ami_deployment_model=ami_deployment_model,
+                                             security_group_ids=security_group_ids,
+                                             network_config_results=network_config_results,
+                                             logger=logger)
 
         return aws_model
 
+    def _prepare_network_interfaces(self, vpc, ami_deployment_model, security_group_ids, network_config_results,
+                                    logger):
+        """
+        :param vpc: The reservation VPC
+        :param DeployAWSEc2AMIInstanceResourceModel ami_deployment_model:
+        :param [str] security_group_ids:
+        :param list[DeployNetworkingResultModel] network_config_results: list of network configuration result objects
+        :param logging.Logger logger:
+        :return:
+        """
+        if ami_deployment_model.network_configurations is None:
+            logger.info("Single subnet mode detected")
+            network_config_results[0].device_index = 0
+            return [self.network_interface_service.get_network_interface_for_single_subnet_mode(
+                    add_public_ip=ami_deployment_model.add_public_ip,
+                    security_group_ids=security_group_ids,
+                    vpc=vpc)]
+
+        self._validate_network_interfaces_request(ami_deployment_model, logger)
+
+        net_interfaces = []
+        public_ip_prop_value = \
+            None if len(ami_deployment_model.network_configurations) > 1 else ami_deployment_model.add_public_ip
+
+        logger.info("Applying device index strategy")
+        self.device_index_strategy.apply(ami_deployment_model.network_configurations)
+
+        logger.info("Building network interface dtos")
+        for net_config in ami_deployment_model.network_configurations:
+            if not isinstance(net_config.connection_params, SubnetConnectionParams):
+                continue
+
+            device_index = net_config.connection_params.device_index
+
+            net_interfaces.append(
+                    # todo: maybe add fallback to find subnet by cidr if subnet id doesnt exist?
+                    self.network_interface_service.build_network_interface_dto(
+                            subnet_id=net_config.connection_params.subnet_id,
+                            device_index=device_index,
+                            groups=security_group_ids,  # todo: set groups by subnet id
+                            public_ip=public_ip_prop_value))
+
+            # set device index on action result object
+            res = first_or_default(network_config_results, lambda x: x.action_id == net_config.id)
+            res.device_index = device_index
+
+        if len(net_interfaces) == 0:
+            logger.info("No network interface dto was created, switching back to single subnet mode")
+            network_config_results[0].device_index = 0
+            return [self.network_interface_service.get_network_interface_for_single_subnet_mode(
+                    add_public_ip=ami_deployment_model.add_public_ip,
+                    security_group_ids=security_group_ids,
+                    vpc=vpc)]
+
+        logger.info("Created dtos for {} network interfaces".format(len(net_interfaces)))
+
+        return net_interfaces
+
+    def _validate_network_interfaces_request(self, ami_deployment_model, logger):
+        self._validate_public_ip_with_multiple_subnets(ami_deployment_model, logger)
+        self._validate_public_subnet_exist_if_requested_public_or_elastic_ips(ami_deployment_model, logger)
+
+    def _validate_public_ip_with_multiple_subnets(self, ami_deployment_model, logger):
+        if ami_deployment_model.add_public_ip and len(ami_deployment_model.network_configurations) > 1:
+            logger.error("Requested public ip with multiple subnets")
+            raise ValueError("Public IP option is not supported with multiple subnets")
+
     def _get_instance_item(self, ami_deployment_model, aws_ec2_resource_model):
         return ami_deployment_model.instance_type if ami_deployment_model.instance_type else aws_ec2_resource_model.instance_type
 
-    def _set_security_group_param(self, aws_model, reservation, security_group, vpc):
-        default_sg_name = self.security_group_service.get_sandbox_security_group_name(reservation.reservation_id)
-        default_sg = self.security_group_service.get_security_group_by_name(vpc, default_sg_name)
+    def _get_security_group_param(self, reservation, security_group, vpc, allow_sandbox_traffic):
+        security_group_ids = []
 
-        aws_model.security_group_ids = [default_sg.id]
+        if allow_sandbox_traffic:
+            default_sg_name = self.security_group_service.sandbox_default_sg_name(reservation.reservation_id)
+        else:
+            default_sg_name = self.security_group_service.sandbox_isolated_sg_name(reservation.reservation_id)
+
+        sg = self.security_group_service.get_security_group_by_name(vpc, default_sg_name)
+        security_group_ids.append(sg.id)
 
         if security_group:
-            aws_model.security_group_ids.append(security_group.group_id)
+            security_group_ids.append(security_group.group_id)
+
+        return security_group_ids
 
     def _get_block_device_mappings(self, image, ami_deployment_model, aws_ec2_resource_model):
         """
@@ -342,26 +533,12 @@ class DeployAMIOperation(object):
         """
         return int(storage_size) * 30
 
-    def _set_elastic_ip(self, ec2_session, ec2_client, instance):
-        """
-        :param ec2_session: EC2 session
-        :param ec2_client: EC2 client
-        :param instance:
-        :return: allocated elastic ip
-        :rtype: str
-        """
-        elastic_ip = self.instance_service.allocate_elastic_address(ec2_client=ec2_client)
-        self.instance_service.associate_elastic_ip(ec2_session=ec2_session,
-                                                   instance=instance,
-                                                   elastic_ip=elastic_ip)
-        return elastic_ip
-
-    def _prepare_deployed_app_attributes(self, instance, ami_credentials, ami_deployment_model, allocated_elastic_ip):
+    def _prepare_deployed_app_attributes(self, ami_credentials, ami_deployment_model, network_config_results):
         """
 
-        :param instance:
         :param cloudshell.cp.aws.models.ami_credentials.AMICredentials ami_credentials:
         :param cloudshell.cp.aws.models.deploy_aws_ec2_ami_instance_resource_model.DeployAWSEc2AMIInstanceResourceModel ami_deployment_model:
+        :param list[DeployNetworkingResultModel] network_config_results:
         :return:
         :rtype: dict
         """
@@ -372,20 +549,21 @@ class DeployAMIOperation(object):
                 deployed_app_attr['Password'] = ami_credentials.password
             deployed_app_attr['User'] = ami_credentials.user_name
 
-        if ami_deployment_model.add_public_ip:
-            deployed_app_attr['Public IP'] = instance.public_ip_address
-        if allocated_elastic_ip:
-            deployed_app_attr['Public IP'] = allocated_elastic_ip
+        if ami_deployment_model.add_public_ip or ami_deployment_model.allocate_elastic_ip:
+            # get the first public ip after sorting the network_config_results by device index
+            deployed_app_attr['Public IP'] = \
+                first_or_default(sorted(network_config_results, key=lambda x: x.device_index),
+                                 lambda x: x.public_ip).public_ip
 
         return deployed_app_attr
 
-    def _rollback_deploy(self, ec2_session, instance_id, custom_security_group, elastic_ip, logger):
+    def _rollback_deploy(self, ec2_session, instance_id, custom_security_group, network_config_results, logger):
         """
 
         :param boto3.ec2.client ec2_session:
         :param str instance_id:
         :param custom_security_group: Security Group object
-        :param str elastic_ip:
+        :param list[DeployNetworkingResultModel] network_config_results:
         :param logging.Logger logger:
         :return:
         """
@@ -402,14 +580,28 @@ class DeployAMIOperation(object):
                                                                            custom_security_group.group_name))
             self.security_group_service.delete_security_group(custom_security_group)
 
-        if elastic_ip:
-            logger.debug("Releasing elastic ip {}".format(elastic_ip))
-            self.instance_service.find_and_release_elastic_address(ec2_session=ec2_session,
-                                                                   elastic_ip=elastic_ip)
+        if network_config_results and len(network_config_results):
+            for r in filter(lambda x: x.public_ip, network_config_results):
+                logger.debug("Releasing elastic ip {}".format(r.public_ip))
+                self.elastic_ip_service.find_and_release_elastic_address(ec2_session=ec2_session,
+                                                                         elastic_ip=r.public_ip)
 
     def _validate_image_available(self, image, ami_id):
         if hasattr(image, 'state') and image.state == 'available':
             return
         raise ValueError('AMI {} not found'.format(ami_id))
 
-
+    def _populate_network_config_results_with_interface_data(self, instance, network_config_results):
+        """
+        :param instance:
+        :param list[DeployNetworkingResultModel] network_config_results:
+        """
+        for interface in instance.network_interfaces_attribute:
+            result = first_or_default(network_config_results,
+                                      lambda x: x.device_index == interface["Attachment"]["DeviceIndex"])
+            result.interface_id = interface["NetworkInterfaceId"]
+            result.private_ip = interface["PrivateIpAddress"]
+            result.mac_address = interface["MacAddress"]
+            if "Association" in interface and "PublicIp" in interface["Association"] \
+                    and interface["Association"]["PublicIp"]:
+                result.public_ip = interface["Association"]["PublicIp"]
